# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import networkx as nx
from .graph_utils import compute_avg_step_length, get_all_predecessors

def set_distance_restraint(molecule,
                           target_node,
                           ref_node,
                           distance,
                           avg_step_length,
                           tolerance):
    """
    Given that a target_node is supposed to be restrained to a reference node at a given
    distance, this function computes for each node in the molecule an upper and lower
    bound references distance that this node needs to have in relation to the target node.
    Those two distances are stored in the 'restraint' attribute of the node.  This information
    is picked up by the `:func:polyply.src.random_walk.checks_milestones` function, which
    then checks that each bound is met by a step in the random_walk.

    The upper_bound is defined as the graph distance of a given node from the target_node
    times the avg_step_length plus the cartesian distance at which the node is
    restrained.

    The lower_bound is defined as the average step-length multiplied by the distance of the
    node from the reference node.

    Parameters
    ----------
    molecule: :class:`vermouth.molecule.Molecule`
    target_node: collections.abc.hashable
        node key of the node that is supposed to be at a
        distance from ref_node
    ref_node: collections.abc.hashable
        node key if this is a distance restraint
    distance: float
        the cartesian distance between nodes in nm
    avg_step_length: float
        average step length (nm)
    tolerance: float
        absolute tolerance (nm)
    """
    # if the target node is a predecssor to the ref node the order
    # needs to be reveresed because the target node will be placed
    # before the ref node
    # we get the path lying between target and reference node
    ancestor = nx.algorithms.lowest_common_ancestor(molecule.search_tree, target_node, ref_node)
    if ancestor == target_node:
        ref_node, target_node = target_node, ref_node
    elif ancestor != ref_node:
        msg=("Your distance restraint between node { } { } is not valid. "
             "Likely you are trying to apply distance restraints on a "
             "branched molecule. This is not fully supported yet. ")
        raise OSError(msg.format(ref_node, target_node))

    path = get_all_predecessors(molecule.search_tree, node=target_node, start_node=ref_node)

    # graph distances can be computed from the path positions
    graph_distances_ref = {node: path.index(node) for node in path}
    graph_distances_target = {node: len(path) - 1 - graph_distances_ref[node] for node in path}

    for node in path:
        if node == target_node:
            graph_distance = 1.0
        elif node == ref_node:
            continue
        else:
            graph_distance = graph_distances_target[node]

        upper_bound = graph_distance * avg_step_length + distance + tolerance

        avg_needed_step_length = distance / graph_distances_target[ref_node]
        lower_bound = avg_needed_step_length * graph_distances_ref[node] - tolerance

        current_restraints = molecule.nodes[node].get('distance_restraints', [])
        molecule.nodes[node]['distance_restraints'] = current_restraints + [(ref_node,
                                                                             upper_bound,
                                                                             lower_bound)]

def set_restraints(topology, nonbond_matrix):
    """
    Set position and/or distance restraints for all molecules.

    topology: `:class:polyply.src.topology.nonbond_matrix`
        the topology of the system
    nonbond_matrix: `:class:polyply.src.NonBondEngine`
        the nonbond matrix which stores all the pairwise interaction
        parameters and positions
    """
    for mol_name, mol_idx in topology.distance_restraints:
        distance_restraints = topology.distance_restraints[(mol_name, mol_idx)]
        mol = topology.molecules[mol_idx]

        for ref_node, target_node in distance_restraints:
            path = list(mol.search_tree.edges)
            avg_step_length, _ = compute_avg_step_length(mol,
                                                         mol_idx,
                                                         nonbond_matrix,
                                                         path)

            distance, tolerance = distance_restraints[(ref_node, target_node)]
            set_distance_restraint(mol, target_node, ref_node, distance, avg_step_length, tolerance)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
High level API for the polyply coordinate generator
"""
import sys
import numpy as np
import networkx as nx
import vermouth.forcefield
from vermouth.file_writer import DeferredFileWriter
from vermouth.log_helpers import StyleAdapter, get_logger
from .generate_templates import GenerateTemplates
from .backmap import Backmap
from .topology import Topology
from .build_system import BuildSystem
from .annotate_ligands import AnnotateLigands, parse_residue_spec, _find_nodes
from .build_file_parser import read_build_file
from .check_residue_equivalence import check_residue_equivalence

LOGGER = StyleAdapter(get_logger(__name__))

def find_starting_node_from_spec(topology, start_nodes):
    """
    Given a definition of one or multiple nodes in
    `start_nodes` find the corresponding molecule index
    and node key in `topology`. The format of start nodes
    is equivalent to that for annotating the residue and
    is described in detail in the function
    :meth:`polyply.src.annotate_ligands.annotate_molecules.parse_residue_spec`.

    Parameters:
    -----------
    topology: :class:`polyply.src.topology.Topology`
    start_nodes: list[str]
        a list of residue-specs

    Returns:
    --------
    dict
        a dict of molecule_idx as key and node key as value
    """
    start_dict = {mol_idx: None for mol_idx in range(len(topology.molecules))}
    for start in start_nodes:
        res_spec = parse_residue_spec(start)
        if 'mol_idx' in res_spec:
            mol_idx = res_spec['mol_idx']
            node = list(_find_nodes(topology.molecules[mol_idx], res_spec))[0]
            start_dict[mol_idx] = node
            topology.molecules[mol_idx].root = node
        else:
            for idx, molecule in enumerate(topology.molecules):
                if molecule.mol_name == res_spec['molname']:
                    node = list(_find_nodes(molecule, res_spec))[0]
                    start_dict[idx] = node
    return start_dict

def _initialize_cylces(topology, cycles, tolerance):
    for mol_name in cycles:
        for mol_idx in topology.mol_idx_by_name[mol_name]:
            # initalize as dfs tree
            molecule = topology.molecules[mol_idx]
            cycles = nx.cycle_basis(molecule)
            if len(cycles) > 1:
                raise IOError("More than one cycle is not allowed.")
            molecule.dfs=True
            nodes = (list(molecule.search_tree.edges)[0][0],
                     list(molecule.search_tree.edges)[-1][1])
            topology.distance_restraints[(mol_name, mol_idx)][nodes] = (0.0, tolerance)

def _check_molecules(molecules):
    """
    Helper method which raises an IOError
    if any molecule in `molecules` is
    disconnected.
    """
    # check if molecules are all connected
    for molecule in molecules:
        if not nx.is_connected(molecule):
            msg = ('\n Molecule {} consistes of two disconnected parts. '
                   'Make sure all atoms/particles in a molecule are '
                   'connected by bonds, constraints or virual-sites')
            raise IOError(msg.format(molecule.name))

def gen_coords(toppath,
               outpath,
               name,
               coordpath=None,
               coordpath_meta=None,
               build=None,
               build_res=[],
               ignore=[],
               cycles=[],
               cycle_tol=0.0,
               split=[],
               ligands=[],
               grid_spacing=0.2,
               grid=None,
               maxiter=800,
               start=[],
               density=None,
               box=None,
               maxiter_random=100,
               step_fudge=1.0,
               max_force=5*10**4.0,
               nrewind=5,
               bfudge=0.4):
    """
    Subprogram for coordinate generation which implements the default
    polyply workflow for structure generation. In general, a topology
    file is read from which all molecules are extracted. Subsequently for each
    residue in the system a template is built. Afterwards a random walk
    is performed for each residue based on a volume estimated from
    the templates. Once the random walk residue coordinates are generated, they
    are backmapped.

    Parameters
    ----------
    toppath: :class:pathlib.Path
        Path to topology file
    outpath: :class:pathlib.Path
        Path to coordinate file
    name: str
        Name of the molecule
    build: :class:pathlib.Path
        Path to build file
    build_res: list[str]
        List of resnames to build
    ignore: list[str]
        List of molecule names to ignore
    cycles: list[str]
        List of cyclic molecule names
    cycle_tol: float
        Tolerance in nm for cycles to be closed
    split: list[str]
        Split single residue into more residues. The syntax is
        <resname>:<new_resname>-<atom1>,<atom2><etc>:<new_resname>
        Note that all atoms of the old residues need to be in at
        most one new residue and must all be accounted for.
    ligands: list[str]
        Specify ligands of molecules which should be placed close
        to a specific molecule. The format is as follows:
        <mol_name>#<mol_idx>-<resname>#<resid>:<mol_name>#<mol_idx>-<resname>#<resid>.
        Elements of the specification can be omitted as required.
        Note mol_name and resname cannot contain the characters
        # and -.
    grid_spacing: float
        Grid spacing in nm
    grid: str
        Path to file with grid-points
    maxiter: int
        Maximum number of tries to generate coordinates for a molecule.
        The default is 800.
    start: list[str]
        Specify which residue to build first. The syntax is
        <mol_name>#<mol_idx>-<resname>#<resid>. Parts of this
        specification may be omitted.
        Note mol_name and resname cannot contain the characters #
        and -.
    density: float
        Density of the system (kg/m3)
    box: np.ndarray(1,3)
        Rectangular box x, y, z in nm
    maxiter_random: int
        Maximum number of tries to place a residue with
        the random walk module.
    step_fudge: float
        Scale the step length in random walk by this fudge factor.
        The default is 1.
    max_force: float
        Maximum force under which placement of a residue is accepted.
        The default is 5x10^4 kJ/(mol*nm).
    nrewind: int
        Number of residues to trace back when random walk step fails in first
        attempt. The default is 5.
    bfudge: float
        Fudge factor by which to scale the coordinates of the residues
        during the backmapping step. 1 will result in to-scale coordinates
        but will likely generate overlaps.
    """

    # Read in the topology
    LOGGER.info("reading topology",  type="step")
    topology = Topology.from_gmx_topfile(name=name, path=toppath)

    LOGGER.info("processing topology",  type="step")
    topology.preprocess()
    _check_molecules(topology.molecules)

    if split:
        LOGGER.info("splitting residues",  type="step")
        for molecule in topology.molecules:
            molecule.split_residue(split)

    # read in coordinates if there are any
    if coordpath:
        LOGGER.info("loading molecule coordinates",  type="step")
        topology.add_positions_from_file(coordpath,
                                         skip_res=build_res,
                                         resolution='mol')

    # read in meta-molecule coordinates of there are any
    if coordpath_meta:
        LOGGER.info("loading meta_molecule coordinates",  type="step")
        topology.add_positions_from_file(coordpath_meta,
                                         skip_res=build_res,
                                         resolution='meta_mol')
        last_node = len(topology.molecules[-1].nodes) - 1
        if topology.molecules[-1].nodes[last_node]["build"]:
            msg = ("You provided metamolecule coordiantes. However, "
                   "there were not enough coordinates for all metamolecule "
                   "residues. Polyply will built the missing residues.")
            LOGGER.warning(msg)

    # load in built file
    if build:
        LOGGER.info("reading build file",  type="step")
        with open(build) as build_file:
            lines = build_file.readlines()
            read_build_file(lines, topology.molecules, topology)

    # collect all starting points for the molecules
    start_dict = find_starting_node_from_spec(topology, start)

    # handle grid input
    if grid:
        LOGGER.info("loading grid",  type="step")
        grid = np.loadtxt(grid)

    # do a sanity check
    LOGGER.info("checking residue integrity",  type="step")
    check_residue_equivalence(topology)
    # Build polymer structure
    LOGGER.info("generating templates",  type="step")
    GenerateTemplates(topology=topology, max_opt=10).run_system(topology)
    LOGGER.info("annotating ligands",  type="step")
    ligand_annotator = AnnotateLigands(topology, ligands)
    ligand_annotator.run_system(topology)
    LOGGER.info("generating system coordinates",  type="step")
    _initialize_cylces(topology, cycles, cycle_tol)
    BuildSystem(topology,
                start_dict=start_dict,
                density=density,
                max_force=max_force,
                grid_spacing=grid_spacing,
                maxiter=maxiter,
                box=box,
                step_fudge=step_fudge,
                ignore=ignore,
                grid=grid,
                cycles=cycles,
                nrewind=nrewind).run_system(topology.molecules)
    ligand_annotator.split_ligands()
    LOGGER.info("backmapping to target resolution",  type="step")
    Backmap(fudge_coords=bfudge).run_system(topology)
    # Write output
    LOGGER.info("writing output",  type="step")
    command = ' '.join(sys.argv)
    system = topology.convert_to_vermouth_system()
    vermouth.gmx.gro.write_gro(system, outpath, precision=7,
                               title=command, box=topology.box)
    DeferredFileWriter().write()



import networkx as nx
from vermouth.molecule import attributes_match

def check_residue_equivalence(topology):
    """
    Check that each residue in all moleculetypes
    is unique.

    Parameters
    ----------
    topology: :class:`polyply.src.topology.Topology`

    Raises
    ------
    IOError
        raise a single error the first time two non-equivalent
        residues have been found
    """
    visited_residues = {}
    resids = {}
    molnames = {}
    for mol_name, mol_idxs in topology.mol_idx_by_name.items():
        # molecules that are included but not used in the topology
        # simply don't get mol_idxs but we need to skip them here
        if not mol_idxs:
            continue
        molecule = topology.molecules[mol_idxs[0]]
        for node in molecule.nodes:
            resname = molecule.nodes[node]["resname"]
            graph = molecule.nodes[node]["graph"]
            if resname in visited_residues:
                if not nx.is_isomorphic(graph, visited_residues[resname]):
                    msg = ("Residue {resname} with resid {residA} in moleculetype {molnameA} is\n"
                           "different from residue {resname} with resid {residB} in moleculetype\n"
                           "{molnameB}. All residues have to be unique in polyply.\n")
                    raise IOError(msg.format(**{"resname": resname,
                                                "molnameB": mol_name,
                                                "molnameA": molnames[resname],
                                                "residA": molecule.nodes[node]["resid"],
                                                "residB": resids[resname]}))
            else:
                visited_residues[resname] = graph
                molnames[resname] = mol_name
                resids[resname] = molecule.nodes[node]["resid"]



# Copyright 2021 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that sequence files are properly read.
"""
from pathlib import Path
import pytest
import networkx as nx
from polyply import TEST_DATA
from polyply.src.meta_molecule import MetaMolecule
from .example_fixtures import example_meta_molecule
from polyply.src.simple_seq_parsers import (_identify_nucleotypes,
                                            _monomers_to_linear_nx_graph,
                                            _parse_plain,
                                            FileFormatError)

@pytest.mark.parametrize('comments, DNA, RNA', (
    # single DNA comment
      (["DNA lorem ipsum"],
       True,
       False
      ),
    # single RNA comment
      (["RNA lorem ipsum"],
       False,
       True
      ),
    # single DNA comment multiple lines
      (["lorem ipsum", "random line DNA", "DNA another line"],
       True,
       False
      ),
    # single RNA comment multiple lines
      (["lorem ipsum", "random line RNA", "RNA another line"],
       False,
       True
      ),
     ))
def test_identify_nucleotypes(comments, DNA, RNA):
    out_DNA, out_RNA = _identify_nucleotypes(comments)
    assert out_DNA == DNA
    assert out_RNA == RNA

@pytest.mark.parametrize('comments', (
    # both DNA and RNA are defined
      ["DNA RNA ipsum", "another line"],
    # neither DNA and RNA are defined
      ["lorem ipsum", "one more line"]
     ))
def test_identify_nucleotypes_fail(comments):
    with pytest.raises(FileFormatError):
        _identify_nucleotypes(comments)

def _node_match(nodeA, nodeB):
    resname = nodeA["resname"] == nodeB["resname"]
    resid = nodeA["resid"] == nodeB["resid"]
    return resname & resid

def test_monomers_to_linear_nx_graph(example_meta_molecule):
    monomers = ["A", "B", "A"]
    seq_graph = _monomers_to_linear_nx_graph(monomers)
    assert nx.is_isomorphic(seq_graph, example_meta_molecule, node_match=_node_match)

@pytest.mark.parametrize('extension, ', (
      "txt",
      "ig",
      "fasta"
     ))
def test_sequence_parses(extension):
    filepath = Path(TEST_DATA + "/simple_seq_files/test."+ extension)
    seq_graph = MetaMolecule.parsers[extension](filepath)
    monomers = ["DA5", "DT", "DC", "DG", "DT", "DA", "DC", "DA", "DT3"]
    ref_graph = _monomers_to_linear_nx_graph(monomers)
    assert nx.is_isomorphic(seq_graph, ref_graph, node_match=_node_match)

def test_ig_cirle():
    filepath = Path(TEST_DATA + "/simple_seq_files/test_circle.ig")
    seq_graph = MetaMolecule.parsers["ig"](filepath)
    monomers = ["DA", "DT", "DC", "DG", "DT", "DA", "DC", "DA", "DT"]
    ref_graph = _monomers_to_linear_nx_graph(monomers)
    ref_graph.add_edge(0, 8)
    assert seq_graph.edges[(0, 8)]["linktype"] == "circle"
    assert nx.is_isomorphic(seq_graph,
                            ref_graph,
                            node_match=_node_match)

def test_ig_termination_fail():
    filepath = Path(TEST_DATA + "/simple_seq_files/test_fail.ig")
    with pytest.raises(FileFormatError):
        seq_graph = MetaMolecule.parsers["ig"](filepath)

@pytest.mark.parametrize('extension, ', (
      "ig",
      "fasta"
     ))
def test_sequence_parses_RNA(extension):
    filepath = Path(TEST_DATA + "/simple_seq_files/test_RNA."+ extension)
    seq_graph = MetaMolecule.parsers[extension](filepath)
    monomers = ["A5", "U", "C", "G", "U", "A", "C", "A", "U3"]
    ref_graph = _monomers_to_linear_nx_graph(monomers)
    assert nx.is_isomorphic(seq_graph, ref_graph, node_match=_node_match)

def test_unkown_nucleotype_error():
    with pytest.raises(IOError):
        lines = ["AABBBCCTG"]
        _parse_plain(lines, DNA=True, RNA=False)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

import textwrap
import pytest
import math
import numpy as np
import vermouth.forcefield
import vermouth.molecule
from vermouth.molecule import Interaction
from vermouth.pdb.pdb import read_pdb
import polyply.src.meta_molecule
from polyply import TEST_DATA
from polyply.src.topology import Topology

class TestTopology:

    @staticmethod
    def test_from_gmx_topfile():
        top = Topology.from_gmx_topfile(TEST_DATA+"/topology_test/system.top", "test")
        assert len(top.molecules) == 1

    @staticmethod
    def test_add_positions_from_gro():
        """
        This test checks if coordinates are supplied at the molecule level that
        coordinates are transfered to the meta_mol.molecule nodes and that the
        meta_molecule nodes have the COG coordinates of the positions. Furthermore,
        the last residue has no coordinates defined. In this case these nodes have
        no coordinates defined and require the build and backmap attribute to be True.
        In all other cases they need to be False.
        """
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
        top.add_positions_from_file(TEST_DATA + "/topology_test/test.gro")
        for node in top.molecules[0].molecule.nodes:
            if node < 14:
                assert "position" in top.molecules[0].molecule.nodes[node].keys()

        for node in top.molecules[0].nodes:
            if node != 2:
                assert "position" in top.molecules[0].nodes[node].keys()
                assert top.molecules[0].nodes[node]["build"] == False
                assert top.molecules[0].nodes[node]["backmap"] == False
            else:
                assert top.molecules[0].nodes[node]["build"] == True
                assert top.molecules[0].nodes[node]["backmap"] == True

    @staticmethod
    def test_add_meta_positions_from_gro():
        """
        This test checks if coordinates are supplied at the meta_molecule level that
        coordinates are transfered to the meta_mol nodes, but not the molecule nodes.
        Furthermore, the last residue has no coordinates defined. In this case these nodes
        have no coordinates defined at either level and require the build and backmap attribute
        to be True. In all other cases the build attribute is False and backmap is True.
        """
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
        top.add_positions_from_file(TEST_DATA + "/topology_test/meta.gro", resolution="meta_mol")
        for node in top.molecules[0].nodes:
            if node != 2:
                assert "position" in top.molecules[0].nodes[node].keys()
                assert top.molecules[0].nodes[node]["build"] == False
                assert top.molecules[0].nodes[node]["backmap"] == True
            else:
                assert top.molecules[0].nodes[node]["build"] == True
                assert top.molecules[0].nodes[node]["backmap"] == True

    @staticmethod
    def test_add_positions_from_pdb():
        """
        This test checks, if coordinates are supplied at the molecule level and no residues
        are missing or skipped, that all molecule nodes have coordinates defined and that
        the meta_molecule positions are defined as well. In addition all build and backmap
        attributes have to be False.
        """
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/pdb.top", "test")
        top.add_positions_from_file(TEST_DATA + "/topology_test/test.pdb")

        pdb_mols = read_pdb(TEST_DATA + "/topology_test/test.pdb")
        for idx, meta_mol in enumerate(top.molecules):
            for node in meta_mol.molecule.nodes:
                ref_pos = pdb_mols[idx].nodes[node]['position']
                assert np.all(ref_pos == meta_mol.molecule.nodes[node]['position'])

        for meta_mol in top.molecules:
            for node in meta_mol.nodes:
                    assert "position" in meta_mol.nodes[node].keys()
                    assert meta_mol.nodes[node]["build"] == False
                    assert meta_mol.nodes[node]["backmap"] == False

    @staticmethod
    def test_add_positions_from_file_fail():
        """
        This test checks if coordinates for a residue at the molecule level
        are incomplete the appropiate error is raised.
        """
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
        with pytest.raises(IOError):
            top.add_positions_from_file(TEST_DATA + "/topology_test/fail.gro")

    @staticmethod
    def test_convert_to_vermouth_system():
        top = Topology.from_gmx_topfile(TEST_DATA + "/topology_test/system.top", "test")
        system = top.convert_to_vermouth_system()
        assert isinstance(system, vermouth.system.System)
        assert len(system.molecules) == 1

    @staticmethod
    @pytest.mark.parametrize('lines, outcome', (
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        """,
        {frozenset(["O", "O"]): {"f": 1,
                               "nb1": 2.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "C"]): {"f": 1,
                               "nb1": 1.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "O"]): {"f": 1,
                               "nb1": 0.0022106496,
                               "nb2": 9.9002500e-07}}
        ),
        ("""
        [ defaults ]
        1.0   3.0    yes  0.5     0.5
        [ atomtypes ]
        C   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        O   O   8      15.99940    -0.500       A    2.96000e-01  8.78640e-01 ; SIG
        """,
        {frozenset(["C", "C"]): {"f": 1,
                               "nb1": 3.75000e-01,
                               "nb2": 4.39320e-01},
         frozenset(["O", "O"]): {"f": 1,
                               "nb1": 2.96000e-01,
                               "nb2": 8.78640e-01},
         frozenset(["O", "C"]): {"f": 1,
                               "nb1": 0.3355,
                               "nb2": 0.6212923022217481}}
        ),
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        [ nonbond_params ]
        C    O    1     2.0     4.0
        """,
        {frozenset(["O", "O"]): {"f": 1,
                               "nb1": 2.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "C"]): {"f": 1,
                               "nb1": 1.7106496e-03,
                               "nb2": 9.9002500e-07},
         frozenset(["C", "O"]): {"f": 1,
                               "nb1": 2.0,
                               "nb2": 4.0}}
        )))
    def test_gen_pairs(lines, outcome):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        top.gen_pairs()
        for atom_pair in outcome:
            assert atom_pair in top.nonbond_params
            nb1 = top.nonbond_params[atom_pair]["nb1"]
            nb2 = top.nonbond_params[atom_pair]["nb2"]
            nb1_ref = outcome[atom_pair]["nb1"]
            nb2_ref = outcome[atom_pair]["nb2"]
            assert math.isclose(nb1, nb1_ref)
            assert math.isclose(nb2, nb2_ref)

    @staticmethod
    @pytest.mark.parametrize('lines, outcome', (
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        #define ga_1  100  250
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CH3   1 test C1 1   0.0 14.0
        2 CH2   1 test C2 2   0.0 12.0
        3 CH3   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 2 ga_1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        ["2", "100", "250"]
        ),
        # different location of define statement
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        #define ga_1  100  250
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CH3   1 test C1 1   0.0 14.0
        2 CH2   1 test C2 2   0.0 12.0
        3 CH3   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 2 ga_1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        ["2", "100", "250"]
        ),
        # two defines for one statement
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        #define ga_1  100
        #define gk_1  250
        [ atomtypes ]
        O       8 0.000 0.000  A   2.7106496e-03  9.9002500e-07
        C       8 0.000 0.000  A   1.7106496e-03  9.9002500e-07
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CH3   1 test C1 1   0.0 14.0
        2 CH2   1 test C2 2   0.0 12.0
        3 CH3   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 2 ga_1 gk_1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        ["2", "100", "250"]
        )
	))
    def test_replace_defines(lines, outcome):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        top.replace_defines()
        assert top.molecules[0].molecule.interactions["angles"][0].parameters == outcome

    @staticmethod
    def test_convert_nonbond_to_sig_eps():
        """
        Simply test if the conversion from C6 C12 to simga epsilon
        is done properly.
        """

        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        top.nonbond_params = {frozenset(["EO", "EO"]):
                             {"nb1":6.44779031E-02 , "nb2": 4.07588234E-04}}
        top.convert_nonbond_to_sig_eps()
        assert math.isclose(top.nonbond_params[frozenset(["EO", "EO"])]["nb1"], 0.43)
        assert math.isclose(top.nonbond_params[frozenset(["EO", "EO"])]["nb2"], 3.4*0.75)

    @staticmethod
    @pytest.mark.parametrize('lines, outcome', (
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ bondtypes ]
        C       C       1       0.1335  502080.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 C   1 test C1 1   0.0 14.0
        2 C   1 test C2 2   0.0 12.0
        [ bonds ]
        1 2  1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={})]}
        ),
        # test three element define
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ angletypes ]
        CE1   CE1	CT2	5	123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CE1   1 test C1 1   0.0 14.0
        2 CE1   1 test C2 2   0.0 12.0
        3 CT2   1 test C3 3   0.0 12.0
        [ angles ]
        1 2  3 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"angles": [Interaction(atoms=(0, 1, 2), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={})]}
        ),
        # test reverse match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ angletypes ]
        CE1    CE2	CT2	5	123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        [ angles ]
        1  2  3 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"angles": [Interaction(atoms=(0, 1, 2), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={})]}
        ),
        # test generic match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        X  CE2  CE1  X    5   123.50	401.664	0.0	0.0
        X  CT2  CE1  X    5   120	400	0.0	0.0
        X  QQQ  QQQ  X    5   150	400	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        2  3  4  5 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={}),
                       Interaction(atoms=(1, 2, 3, 4), parameters=["5", "120", "400", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test generic match plus defined match on same pattern
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        X  CE2  CE1  X    5   123.50	401.664	0.0	0.0
        X  CT2  CE1  X    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        2  3  4  5 1   150  60  0.0 0.0
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={}),
                       Interaction(atoms=(1, 2, 3, 4), parameters=["1", "150", "60", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test priority of defined over generic match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        CT2  CE2  CE1  CT2    5   123.50	401.664	0.0	0.0
        X    CE2  CE1  X      5   20            20      0.0     0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test reverse order for priority of defined over generic match
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        X    CE2  CE1  X      5   20            20      0.0     0.0
        CT2  CE2  CE1  CT2    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                   meta={})]}
        ),
        # test generic improper
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        CT2   X     X     CT2    5   123.50	401.664	0.0	0.0
        CT2   X     X     CE2    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        5 CT2   1 test C5 5   0.0 14.0
        [ dihedrals ]
        1  2  3  4 2
        2  3  4  5 2
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"dihedrals": [Interaction(atoms=(0, 1, 2, 3), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={}),
                       Interaction(atoms=(1, 2, 3, 4), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={})]}
        ),
        # multiple matchs and pairs and a meta parameters
        (
        """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ angletypes ]
        CE1    CE2	CT2	5	123.50	401.664	0.0	0.0
        [ bondtypes ]
        CT2       CE2       1       0.1335  502080.0
        #ifdef old
        CE2       CE1       1       0.1335  502080.0
        #endif
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        [ bonds ]
        1 2 1
        2 3 1
        [ pairs ]
        1  2  1
        [ angles ]
        #ifdef angle
        1  2  3 1
        #endif
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(1, 2), parameters=["1", "0.1335", "502080.0"], meta={'tag': 'old', 'condition': 'ifdef'})],
         "pairs": [Interaction(atoms=(0, 1), parameters=["1"], meta={})],
         "angles": [Interaction(atoms=(0, 1, 2), parameters=["5", "123.50", "401.664", "0.0", "0.0"],
                                meta={"ifdef":"angle"})]}
        ),
        # test bondtype usage
        (
        """
        #define _FF_OPLS
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        opls_001   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        [ bondtypes ]
        C       C       1       0.1335  502080.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 opls_001   1 test C1 1   0.0 14.0
        2 opls_001   1 test C2 2   0.0 12.0
        [ bonds ]
        1 2  1
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={})]}
        ),
        # test virtual_sites n,2,3,4 are skipped
        (
        """
        #define _FF_OPLS
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ atomtypes ]
        opls_001   C   6      12.01100     0.500       A    3.75000e-01  4.39320e-01 ; SIG
        [ bondtypes ]
        C       C       1       0.1335  502080.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 opls_001   1 test C1 1   0.0 14.0
        2 opls_001   1 test C2 2   0.0 12.0
        3 opls_001   1 test C2 2   0.0 12.0
        4 opls_001   1 test C2 2   0.0 12.0
        5 opls_001   1 test C2 2   0.0 12.0
        6 opls_001   1 test C2 2   0.0 12.0
        7 opls_001   1 test C2 2   0.0 12.0
        8 opls_001   1 test C2 2   0.0 12.0
        9 opls_001   1 test C2 2   0.0 12.0
        10 opls_001   1 test C2 2   0.0 12.0
        [ bonds ]
        1  2  1
        1  8  1
        1  9  1
        2  9  1
        8  9  1
        ; currently not parsed accurately due to vermouth bug
        [ virtual_sites2 ]
        4   1  9  1  0.5000
        [ virtual_sites3 ]
        5   4  8  1  1  0.200  0.200
        6   4  9  2  1  0.200  0.200
        [ virtual_sites4 ]
        10   4  8  1  7  1  0.200  0.200  0.300
        [ virtual_sitesn ]
        3   1   4   4   1  2
        7   1   4   4   8  9
        [ system ]
        some title
        [ molecules ]
        test 1
        """,
        {"bonds": [Interaction(atoms=(0, 1), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(0, 7), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(0, 8), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(1, 8), parameters=["1", "0.1335", "502080.0"], meta={}),
                   Interaction(atoms=(7, 8), parameters=["1", "0.1335", "502080.0"], meta={})],
         "virtual_sitesn": [Interaction(atoms=(2, 3, 3, 0, 1), parameters=["1"], meta={}),
                            Interaction(atoms=(6, 3, 3, 7, 8), parameters=["1"], meta={})],
         "virtual_sites4": [Interaction(atoms=(9, 3, 7, 0, 6), parameters=["1", "0.200", "0.200", "0.300"], meta={})],
         "virtual_sites2": [Interaction(atoms=(3, 0, 8), parameters=["1", "0.5000"], meta={})],
         "virtual_sites3": [Interaction(atoms=(4, 3, 7, 0), parameters=["1", "0.200", "0.200"], meta={}),
                            Interaction(atoms=(5, 3, 8, 1), parameters=["1", "0.200", "0.200"], meta={})]}
        )
	))
    def test_replace_types(lines, outcome):
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        top.gen_bonded_interactions()
        for inter_type in outcome:
            assert top.molecules[0].molecule.interactions[inter_type] == outcome[inter_type]

    @staticmethod
    def test_replace_types_fail():
        lines = """
        [ defaults ]
        1.0   1.0   yes  1.0     1.0
        [ dihedraltypes ]
        C   Q    Q     Q    5   123.50	401.664	0.0	0.0
        [ moleculetype ]
        test 3
        [ atoms ]
        1 CT2   1 test C1 1   0.0 14.0
        2 CE2   1 test C2 2   0.0 12.0
        3 CE1   1 test C3 3   0.0 12.0
        4 CT2   1 test C4 4   0.0 12.0
        [ dihedrals ]
        1  2  3  4 2
        [ system ]
        some title
        [ molecules ]
        test 1
        """
        new_lines = textwrap.dedent(lines)
        new_lines = new_lines.splitlines()
        force_field = vermouth.forcefield.ForceField(name='test_ff')
        top =  Topology(force_field, name="test")
        polyply.src.top_parser.read_topology(new_lines, top)
        with pytest.raises(OSError):
             top.gen_bonded_interactions()



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that distance and position restraints are tagged.
"""
import pytest
import numpy as np
import networkx as nx
import polyply
from polyply.tests.test_build_file_parser import test_molecule

@pytest.mark.parametrize('target_node, ref_node, distance, avg_step_length, tolerance, expected',(
    # test simple revers
   (0, 4, 1.5, 0.47, 0.0,
    {1: [(0, 2.91, 0.375)],
     2: [(0, 2.44, 0.75)],
     3: [(0, 1.97, 1.125)],
     4: [(0, 1.97, 1.5)],}
   ),
    # test simple case
   (4, 0, 1.5, 0.47, 0.0,
    {1: [(0, 2.91, 0.375)],
     2: [(0, 2.44, 0.75)],
     3: [(0, 1.97, 1.125)],
     4: [(0, 1.97, 1.5)],
    }
   ),
   (4, 0, 1.5, 0.47, 0.3,
    {1: [(0, 3.21, 0.075)],
     2: [(0, 2.74, 0.45)],
     3: [(0, 2.27, 0.825)],
     4: [(0, 2.27, 1.2)],
    }
   ),
   ))
def test_set_distance_restraint(test_molecule,
                                target_node,
                                ref_node,
                                distance,
                                avg_step_length,
                                tolerance,
                                expected):

    polyply.src.restraints.set_distance_restraint(test_molecule,
                                                  target_node,
                                                  ref_node,
                                                  distance,
                                                  avg_step_length,
                                                  tolerance)

    attr_list = nx.get_node_attributes(test_molecule, "distance_restraints")
    for node, restr_list in attr_list.items():
        ref_list = expected[node]
        for ref_restr, new_restr in zip(ref_list, restr_list):
            assert ref_restr[0] == new_restr[0]
            assert np.isclose(ref_restr[1], new_restr[1], atol=0.001)
            assert np.isclose(ref_restr[2], new_restr[2], atol=0.001)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This collection of tests checks some basic I/O workflows of the gen_coords
function. For example, that it runs through when no coordinates are requested
to be generated.
"""
import logging
import pathlib
import pytest
import numpy as np
import networkx as nx
from vermouth.gmx.gro import read_gro
from vermouth.graph_utils import make_residue_graph
import polyply
from polyply import TEST_DATA, gen_coords

def test_no_positions_generated(tmp_path, monkeypatch):
    """
    All positions are defined none have to be created. Should run through without
    errors and preserve all positions given in the input to rounding accuracy.
    """
    monkeypatch.chdir(tmp_path)
    top_file = TEST_DATA + "/topology_test/system.top"
    pos_file = TEST_DATA + "/topology_test/complete.gro"
    out_file = tmp_path / "out.gro"
    gen_coords(toppath=top_file,
               coordpath=pos_file,
               outpath=out_file,
               name="test",
               box=np.array([10, 10, 10])
               )
    molecule_out = read_gro(out_file, exclude=())
    molecule_in = read_gro(pos_file, exclude=())
    for node in molecule_out.nodes:
        assert np.all(molecule_out.nodes[node]['position'] ==
                      molecule_in.nodes[node]['position'])

def test_backmap_only(tmp_path, monkeypatch):
    """
    Only meta_mol positions are defined so others have to be backmapped.
    Should run through without errors and COG of residues should be the
    same as they have been put in.
    """
    monkeypatch.chdir(tmp_path)
    top_file = TEST_DATA + "/topology_test/system.top"
    pos_file = TEST_DATA + "/topology_test/cog.gro"
    out_file = tmp_path / "out.gro"
    gen_coords(toppath=top_file,
               coordpath_meta=pos_file,
               outpath=out_file,
               name="test",
               box=np.array([11, 11, 11])
               )
    molecule_out = read_gro(out_file, exclude=())
    for node in molecule_out.nodes:
        assert "position" in molecule_out.nodes[node]

    meta_in = read_gro(pos_file, exclude=())
    meta_out = make_residue_graph(molecule_out)
    for node in meta_out.nodes:
        nodes_pos = nx.get_node_attributes(meta_out.nodes[node]['graph'], 'position')
        nodes_pos = np.array(list(nodes_pos.values()))
        res_pos = np.average(nodes_pos, axis=0)
        ref_pos = meta_in.nodes[node]['position']
        # tolerance comes from finite tolerance in gro file
        assert np.allclose(res_pos, ref_pos, atol=0.0009)

def test_warning_partial_metamol_coords(tmp_path, monkeypatch, caplog):
    caplog.set_level(logging.WARNING)
    top_file = TEST_DATA + "/topology_test/system.top"
    pos_file = TEST_DATA + "/topology_test/cog_missing.gro"
    out_file = tmp_path / "out.gro"

    with caplog.at_level(logging.WARNING):
        gen_coords(toppath=top_file,
                   coordpath_meta=pos_file,
                   outpath=out_file,
                   name="test",
                   box=np.array([11, 11, 11])
                   )
        print(caplog.records)
        for record in caplog.records:
            assert record.levelname == "WARNING"
            break
        else:
            assert False
